require "src/vector"
require "src/triangle"
require "src/mesh"
require "src/matrix"

function love.load()
    ---------------------------setup screen size-----------------------------------------
    screen = {}
    screen.dpiScale = love.graphics.getDPIScale()
    screen.width, screen.height = love.window.getMode()
    screen.width, screen.height = screen.width / screen.dpiScale, screen.height / screen.dpiScale
    screen.width_half = screen.width / 2
    screen.height_half = screen.height / 2

    love.graphics.setLineJoin("none")
    Meshes = {}
    mesh = {}
    mesh.triangles =  LoadMesh("meshes/teapot.obj")
    mesh.name = "Enterprise"
    mesh.pos = Vector.new(0, 0.0, 0)
    table.insert(Meshes, mesh)
    --mesh = {}
    --mesh.triangles = LoadMesh("meshes/saucer.obj")
    --mesh.pos = Vector.new(0, 0, 0)
    --mesh.name = "Saucer"


    --table.insert(Meshes, mesh)

    for g, mesh in ipairs(Meshes) do 
        Move_mesh(Meshes[g])
    end

    Camera = Vector.new(0, 0, 0)
    Camera_speed = 1
    Camera_AngleY = 0
    Camera_AngleX = 0


    LookDir = Vector.new(0, 0, 0)
    Light_dir = Vector.new(1, 1, -1)
    LookDir = Vector.new(0, 0, 1)
    stop, start = 0, 0
    love.mouse.setGrabbed(true)
    love.mouse.setRelativeMode(true)
    MousePos = {}
    Mouse_sensitivity = math.pi / 4


    Theta = 0
    MatProj = Matrix.createProjection(90, 1000, 0.1, screen.width, screen.height)
    MatRotX = Matrix.createRotationX(Theta)
    MatRotZ = Matrix.createRotationZ(Theta)
    MatTrans = Matrix.createTranslate(0, 0, 2)
    MatWorld = Matrix.new(4, 4)
    MatWorld = Matrix.multiplyByMatrix(MatRotZ, MatRotX)
    MatWorld = Matrix.multiplyByMatrix(MatWorld, MatTrans)

    RotZ = 0
    RotX = 0

    Stars = Matrix.new(screen.width, screen.height)
    for r=1, screen.width do 
        for c = 1,  screen.height do 
            Stars.m[r][c] = math.random()
        end
    end
end

function love.update(dt) 
    
    for g, mesh in ipairs(Meshes) do 
        if Meshes[g].name == "Enterprise" then 
            Meshes[g].pos = Control_Mesh(mesh) 
        end
        if Vector.length(Meshes[g].pos) > 0 then 

            Move_mesh(Meshes[g])
        end
        Meshes[g].pos = Vector.new(0, 0, 0)
    end
    love.window.setTitle(love.timer.getFPS())
    Camera_move(dt)
    MatRotZ = Matrix.createRotationZ(Theta)
    MatRotX = Matrix.createRotationX(Theta)
    MatTrans = Matrix.createTranslate(0, 0, 2)
    MatWorld = Matrix.multiplyByMatrix(MatRotZ, MatRotX)
    MatWorld = Matrix.multiplyByMatrix(MatWorld, MatTrans)


    if love.keyboard.isDown("r") then 
        Camera = Vector.new(0, 0, 0)
        Camera_speed = 3
        Camera_AngleY = 0
        Camera_AngleX = 0
    end

    local up = Vector.new(0, 1, 0)
    local target = Vector.new(0, 0, 1)
    MatCamRotY = Matrix.createRotationY(Camera_AngleY)
    MatCamRotX = Matrix.createRotationX(Camera_AngleX)
    MatCamRotAll = Matrix.multiplyByMatrix(MatCamRotX, MatCamRotY)

   


    LookDir = Vector.multiplyByMatrix(target, MatCamRotAll)
    target = Vector.add(Camera, LookDir)
    MatCamera = Matrix.pointAt(Camera, target, up)
    MatView = Matrix.invert(MatCamera)
    Maths()
end

function Move_mesh(mesh)
    for i, tris in ipairs(mesh.triangles) do
        tris.vectors[1] = Vector.add(tris.vectors[1], mesh.pos)
        tris.vectors[2] = Vector.add(tris.vectors[2], mesh.pos)
        tris.vectors[3] = Vector.add(tris.vectors[3], mesh.pos)
    end
end


function love.draw()
    for i=1, #list do 
        love.graphics.setColor(list[i].col[1], list[i].col[2], list[i].col[3], 1) 
       love.graphics.polygon("fill",
       list[i].vectors[1].x, list[i].vectors[1].y,
       list[i].vectors[2].x, list[i].vectors[2].y, 
       list[i].vectors[3].x,list[i].vectors[3].y)
       --love.graphics.setColor(1, 1, 1, 1)
       --love.graphics.polygon("line",
       --list[i].vectors[1].x, list[i].vectors[1].y,
       --list[i].vectors[2].x, list[i].vectors[2].y, 
       --list[i].vectors[3].x,list[i].vectors[3].y)
    end
 
    love.graphics.print("Frametime:" .. stop - start)
end

function  Control_Mesh(mesh)
    if love.keyboard.isDown("down") then 
        mesh.pos.x = mesh.pos.x + 0.01 
        
    elseif love.keyboard.isDown("up") then 
        mesh.pos.x = mesh.pos.x - 0.01 
        --if mesh.pos.z > 0 then mesh.pos.z = mesh.pos.z / 2 * dt end
    end
    return mesh.pos
end

local PainterAlg = function ( a, b )
    return (a.vectors[1].z + a.vectors[2].z + a.vectors[3].z)/3 > (b.vectors[1].z + b.vectors[2].z + b.vectors[3].z)/3
  end

function Sort_Triangles(list)
    table.sort (list, PainterAlg)
   
end
--fix strafing when looking down
function  Camera_move(dt)
    local forward = Vector.mul(LookDir, Camera_speed * dt)

    if love.keyboard.isDown("space") then 
        Camera.y = Camera.y + Camera_speed * dt
    end
    if love.keyboard.isDown("lshift") then 
        Camera.y = Camera.y - Camera_speed * dt
    end
    if love.keyboard.isDown("a") then 
        Camera = Vector.subtract(Camera, Vector.crossProduct(forward, Vector.new(0, 1, 0)))
    end
    if love.keyboard.isDown("d") then 
        Camera = Vector.add(Camera, Vector.crossProduct(forward, Vector.new(0, 1, 0)))

    end
    if love.keyboard.isDown("w") then 
        Camera = Vector.add(Camera, forward)
    end
    if love.keyboard.isDown("s") then 
        Camera = Vector.subtract(Camera, forward)

    end
  
end 

function love.mousemoved(x, y, dirx, diry, touch )
    dt = love.timer.getAverageDelta()
    if dirx < 0  then 
        Camera_AngleY = Camera_AngleY - Mouse_sensitivity * dt
    elseif dirx > 0 then
        Camera_AngleY = Camera_AngleY + Mouse_sensitivity * dt
    end
    if diry < 0  then 
        Camera_AngleX = Camera_AngleX - Mouse_sensitivity * dt
    elseif diry > 0 then
        Camera_AngleX = Camera_AngleX + Mouse_sensitivity * dt

    end
    --prevent the camera from looping back around/breaking your neck
    if Camera_AngleX > math.pi-0.1 then Camera_AngleX = math.pi-0.1 end
    if Camera_AngleX < -math.pi+0.1 then Camera_AngleX = -math.pi+0.1 end

end

function Maths()
    list = {}
    start = love.timer.getTime()
    for g, mesh in ipairs(Meshes) do
        for i, tris in ipairs(Meshes[g].triangles) do
            local triProjected = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )
            local triTransformed = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )
            local triViewed = Triangle.new({Vector.new(0,0,0), Vector.new(0,0,0), Vector.new(0,0,0)} )
            --rotate in x and z and push into v
            
            tris.vectors[1].w = 1
            tris.vectors[2].w = 1
            tris.vectors[3].w = 1


            triTransformed.vectors[1] = Vector.multiplyByMatrix(tris.vectors[1], MatWorld)
            triTransformed.vectors[2] = Vector.multiplyByMatrix(tris.vectors[2], MatWorld)
            triTransformed.vectors[3] = Vector.multiplyByMatrix(tris.vectors[3], MatWorld)


            --culling
            local normal = Vector.new(0, 0, 0); local line1 = Vector.new(0, 0, 0); local line2 = Vector.new(0, 0, 0);
            line1 = Vector.subtract(triTransformed.vectors[2], triTransformed.vectors[1])
            line2 = Vector.subtract(triTransformed.vectors[3], triTransformed.vectors[1])
            normal = Vector.crossProduct(line1, line2)
            normal = Vector.norm(normal)

            local cameraRay =  Vector.subtract( triTransformed.vectors[1], Camera) 

            if Vector.dotProduct(normal, cameraRay) < 0 then
                --lighting
                Light_dir = Vector.norm(Light_dir)
                local shadow =  math.max(0.1, Vector.dotProduct(Light_dir, normal))
                triViewed.col = {shadow, shadow, shadow}

                ----world space to view space
                triViewed.vectors[1] = Vector.multiplyByMatrix(triTransformed.vectors[1], MatView)
                triViewed.vectors[2] = Vector.multiplyByMatrix(triTransformed.vectors[2], MatView)
                triViewed.vectors[3] = Vector.multiplyByMatrix(triTransformed.vectors[3], MatView)

                clippedTris = {}
                local result = Triangle.clipAgainstPlane({x = 0, y = 0, z = 0.01}, {x = 0, y = 0, z = 1}, triViewed)
                if result[1] == 1 then table.insert(clippedTris, triViewed) end
                if result[1] == 2 then table.insert(clippedTris, result[2]) end
                if result[1] == 3 then clippedTris[#clippedTris+1] =  result[2] clippedTris[#clippedTris+1] =  result[3] end

                for i =1, #clippedTris do 
                    --Projection from 3D to 2D
                    triProjected.vectors[1] = Vector.multiplyByMatrix( clippedTris[i].vectors[1], MatProj)
                    triProjected.vectors[2] = Vector.multiplyByMatrix( clippedTris[i].vectors[2], MatProj)
                    triProjected.vectors[3] = Vector.multiplyByMatrix( clippedTris[i].vectors[3], MatProj)
                    triProjected.col= clippedTris[i].col
                    --normalize the vectors
                    triProjected.vectors[1] = Vector.div(triProjected.vectors[1], triProjected.vectors[1].w)
                    triProjected.vectors[2] = Vector.div(triProjected.vectors[2], triProjected.vectors[2].w)
                    triProjected.vectors[3] = Vector.div(triProjected.vectors[3], triProjected.vectors[3].w)



                    triProjected.vectors[1].x  = triProjected.vectors[1].x *-1 
	    		    triProjected.vectors[1+1].x = triProjected.vectors[1+1].x  *-1
	    		    triProjected.vectors[2+1].x = triProjected.vectors[2+1].x  *-1
	    		    triProjected.vectors[0+1].y = triProjected.vectors[0+1].y  *-1
	    		    triProjected.vectors[1+1].y = triProjected.vectors[1+1].y  *-1
	    		    triProjected.vectors[2+1].y = triProjected.vectors[2+1].y  *-1

                    --offset into visible normalized space
                    OffsetView = Vector.new(1, 1, 0)
                    triProjected.vectors[1] = Vector.add(triProjected.vectors[1], OffsetView)
                    triProjected.vectors[2] = Vector.add(triProjected.vectors[2], OffsetView)
                    triProjected.vectors[3] = Vector.add(triProjected.vectors[3], OffsetView)
                    triProjected.vectors[1].x = triProjected.vectors[1].x * 0.5 * screen.width;
                    triProjected.vectors[1].y = triProjected.vectors[1].y * 0.5 * screen.height;
                    triProjected.vectors[2].x = triProjected.vectors[2].x * 0.5 * screen.width;
                    triProjected.vectors[2].y = triProjected.vectors[2].y * 0.5 * screen.height;
                    triProjected.vectors[3].x = triProjected.vectors[3].x * 0.5 * screen.width;
                    triProjected.vectors[3].y = triProjected.vectors[3].y * 0.5 * screen.height;

                    ----------------------------inserts screen_space triangles to be sorted
                    table.insert(list,triProjected)
                end
            end
        end
    end
    table.sort (list, PainterAlg)
    stop = love.timer.getTime()
end

function LoadMesh(filename)
    local triangles = {}
    local lines = {}
    local start = love.timer.getTime()
    for line in love.filesystem.lines(filename) do 
        lines[#lines + 1] = line
    end
    for i=1, 3 do --get rid of unnessecary header
         table.remove(lines, 1)
    end
    --load vector
    local j = 1
    local vector = {}
    local c = 0
    local stop = false
    while stop == false do 
        vector[j] = {}
        for num in string.gmatch(lines[j],"%S+") do
            if num ~= "v"  then
                if num == "s" then 
                    stop = true
                end
                if stop == false then 
                    if c + 1 < 4 then 
                        c = c + 1
                    else c = 1 end
                    if c == 1 then 
                        vector[j].x = tonumber(num)
                    elseif c == 2 then 
                        vector[j].y = tonumber(num)
                    elseif c == 3 then 
                        vector[j].z = tonumber(num)
                        vector[j].w = 0
                    end  
                            
                end
            end
        end
        if stop == false then 
            j = j+1
        end
    end
    --put together vector into traingles / faces
    local m = {}
    c=1
    for i=#vector+1, #lines do 
        for num in string.gmatch(lines[i],"%S+") do
            if num ~= "f"  then 
                table.insert( m,tonumber( num))
            end
        end
    end
    for i=1, #lines-#vector do 

        triangles[i] = Triangle.new({vector[m[c]], vector[m[c+1]], vector[m[c+2]]})
        c = c +3 
    end
    return triangles
end
