Matrix = {}

function Matrix.new(rows, col)
    local ins = {}
    ins.m = {}
    for i=1, rows do 
        ins.m[i] = {}
        for j=1, col do 
            ins.m[i][j] = 0
        end
    end
    return ins 
end

function Matrix.createProjection(fov, far, near, scr_width, scr_height)
    local aspectRatio = scr_height/scr_width
    local fovRad =  1 / math.tan(fov*0.5/180* math.pi)
    local m = Matrix.new(4, 4)
    m.m[1][1] = aspectRatio * fovRad
    m.m[2][2] = fovRad
    m.m[3][3] = far / (far - near)
    m.m[4][3] = (-far * near) / (far - near)
    m.m[3][4] = 1
    m.m[4][4] = 0
    return m
end
function Matrix.createRotationX(theta)
    local m = Matrix.new(4, 4)
    m.m[1][1] = 1 
    m.m[2][2] = math.cos(theta* 0.5)
    m.m[2][3] = math.sin(theta* 0.5)
    m.m[3][2] = -math.sin(theta* 0.5)
    m.m[3][3] = math.cos(theta * 0.5)
    m.m[4][4] = 1
    return m
end

function Matrix.createRotationY(theta)
    local m = Matrix.new(4, 4)
    m.m[1][1] = math.cos(theta);
    m.m[1][3] = math.sin(theta);
    m.m[3][1] = -math.sin(theta);
    m.m[2][2] = 1.0
    m.m[3][3] = math.cos(theta);
    m.m[4][4] = 1.0
    return m;
end
function Matrix.createRotationZ(theta)
    local m = Matrix.new(4, 4)
    m.m[1][1] = math.cos(theta)
    m.m[1][2] = math.sin(theta)
    m.m[2][1] = -math.sin(theta)
    m.m[2][2] = math.cos(theta)
    m.m[3][3] = 1
    m.m[4][4] = 1
    return m
end

function Matrix.createTranslate(x, y, z)
    local m = Matrix.new(4, 4)
    m.m[1][1] = 1
    m.m[2][2] = 1
    m.m[3][3] = 1
    m.m[4][4] = 1
    m.m[4][1] = x
    m.m[4][2] = y
    m.m[4][3] = z
    return m
end

function Matrix.multiplyByMatrix(m1, m2)
    local m = Matrix.new(4, 4)
    for r=1, 4 do 
        for c = 1, 4 do 
            m.m[r][c] = m1.m[r][1] * m2.m[1][c] + m1.m[r][2] * m2.m[2][c] + m1.m[r][3] * m2.m[3][c] + m1.m[r][4] * m2.m[4][c]
        end
    end
    return m
end

function Matrix.pointAt(pos, target, up)
    --calculate new forward
    local newForward = Vector.subtract(target, pos)
    newForward = Vector.norm(newForward)
    --calculate new up 
    local a = Vector.mul(newForward, Vector.dotProduct(up, newForward))
    local newUp = Vector.subtract(up, a) 
    newUp = Vector.norm(newUp)

    --newright is cross product
    local newRight = Vector.crossProduct(newUp, newForward)

    --construct Dimensioning and Translation Matrix
    local matrix = Matrix.new(4, 4)
    matrix.m[1][1] = newRight.x;    matrix.m[1][2] = newRight.y;     matrix.m[1][3] = newRight.z     matrix.m[1][4] = 0
    matrix.m[2][1] = newUp.x;       matrix.m[2][2] = newUp.y;        matrix.m[2][3] = newUp.z        matrix.m[2][4] = 0
    matrix.m[3][1] = newForward.x;  matrix.m[3][2] = newForward.y    matrix.m[3][3] = newForward.z   matrix.m[3][4] = 0
    matrix.m[4][1] = pos.x          matrix.m[4][2] = pos.y           matrix.m[4][3] = pos.z          matrix.m[4][4] = 1
    return matrix
end

function  Matrix.invert(m)
    local matrix = Matrix.new(4,4)
	matrix.m[1][1] =   m.m[1][1];  matrix.m[1][2] = m.m[2][1];  matrix.m[1][3] = m.m[3][1];  matrix.m[1][4] = 0
	matrix.m[2][1] =   m.m[1][2];  matrix.m[2][2] = m.m[2][2];  matrix.m[2][3] = m.m[3][2];  matrix.m[2][4] = 0
	matrix.m[3][1] =   m.m[1][3];  matrix.m[3][2] = m.m[2][3];  matrix.m[3][3] = m.m[3][3];  matrix.m[3][4] = 0
	matrix.m[4][1] = -(m.m[4][1] * matrix.m[1][1] + m.m[4][2] * matrix.m[2][1] + m.m[4][3] * matrix.m[3][1]);
	matrix.m[4][2] = -(m.m[4][1] * matrix.m[1][2] + m.m[4][2] * matrix.m[2][2] + m.m[4][3] * matrix.m[3][2]);
	matrix.m[4][3] = -(m.m[4][1] * matrix.m[1][3] + m.m[4][2] * matrix.m[2][3] + m.m[4][3] * matrix.m[3][3]);
	matrix.m[4][4] = 1.0
	return matrix;
end